@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 Objects used to store and manage metabolomics data
 
@@ -25,6 +24,7 @@ constructor:
   party software (XCMS, MZMine2, etc...)
 - Performing Feature correspondence algorithm on features detected from raw
   data (not implemented yet...)
+
 """
 
 from . import utils
@@ -44,14 +44,6 @@ from bokeh.models import LabelSet
 import seaborn as sns
 
 
-# TODO: remove data_path attribute. check with webapp example.
-# TODO: maybe its a good idea to combine export methods into an ExportMethods
-#       object
-# TODO: export DataContainer to metaboanalyst format.
-# TODO: include split into dev / control in DataContainer. This should be
-#  done before any kind of data curation to prevent feature leakage.
-# TODO: make sample metadata compatible with ISA format.
-
 class DataContainer(object):
     """
     Container object that stores processed metabolomics data.
@@ -59,7 +51,7 @@ class DataContainer(object):
     The data is separated in three attributes: data_matrix, sample_metadata and
     feature_metadata. Each one is a pandas DataFrame. DataContainers can be
     created, apart from using the constructor, importing data in common formats
-    (such as: XCMS, MZmine2, Progenesis, etc..) static methods.
+    (such as: XCMS, MZMine2, Progenesis, etc..) static methods.
 
     Attributes
     ----------
@@ -89,7 +81,6 @@ class DataContainer(object):
     metrics : methods to compute common feature metrics.
     plot : methods to plot features.
     preprocess : methods to perform common preprocessing tasks.
-    data_path
     id
     batch
     order
@@ -117,12 +108,12 @@ class DataContainer(object):
     MetricMethods
     PlotMethods
     PreprocessMethods
+
     """
 
     def __init__(self, data_matrix: pd.DataFrame,
                  feature_metadata: pd.DataFrame,
                  sample_metadata: pd.DataFrame,
-                 data_path: Optional[str] = None,
                  mapping: Optional[dict] = None,
                  plot_mode: str = "bokeh"):
         
@@ -139,15 +130,14 @@ class DataContainer(object):
         feature_metadata : pandas.DataFrame.
             DataFrame with features names as indices. mz and rt are required
             columns.
-        data_path : str.
-            path to raw Data. Files must have the same name as each sample.
         mapping : dict or None
             if dict, set each sample class to sample type.
         plot_mode : {"seaborn", "bokeh"}
             The package used to generate plots with the plot methods
+
         """
         validation.validate_data_container(data_matrix, feature_metadata,
-                                           sample_metadata, data_path)
+                                           sample_metadata)
 
         # check and convert order and batch information
         try:
@@ -170,7 +160,6 @@ class DataContainer(object):
         self.sample_metadata = sample_metadata.copy()
         self._sample_mask = data_matrix.index.copy()
         self._feature_mask = data_matrix.columns.copy()
-        self.data_path = data_path
         self.mapping = mapping
         self.id = data_matrix.index
         self.plot = None
@@ -184,25 +173,6 @@ class DataContainer(object):
         self.metrics = MetricMethods(self)
         self.preprocess = PreprocessMethods(self)
         self.set_plot_mode(plot_mode)
-
-    @property
-    def data_path(self) -> pd.DataFrame:
-        """str : path where raw data is stored."""
-        return self._data_path
-    
-    @data_path.setter
-    def data_path(self, path: str):
-        """
-        sets raw data path, search for available samples and adds them to
-        sample information.
-        """
-        if path is not None:
-            path_mapping = utils.sample_to_path(self.data_matrix.index, path)
-            self._sample_metadata.loc[self._sample_mask, _raw_path] = \
-                self.sample_metadata.index.map(path_mapping)
-            self._data_path = path
-        else:
-            self._data_path = None
     
     @property
     def data_matrix(self) -> pd.DataFrame:
@@ -292,18 +262,18 @@ class DataContainer(object):
             msg = "order values must be unique"
             raise ValueError(msg)
 
-    def get_available_samples(self) -> pd.Series:
-        """
-        Returns the absolute path for each raw data file present in
-        data_path.
-        
-        Returns
-        -------
-        available_samples : pd.Series
-            Pandas series with absolute path for each available file.
-        """
-        available_samples = self.sample_metadata[_raw_path].dropna()
-        return available_samples
+    @property
+    def dilution(self) -> pd.Series:
+        try:
+            return self._sample_metadata.loc[self._sample_mask,
+                                             _sample_dilution]
+        except KeyError:
+            msg = "No dilution information available."
+            raise DilutionInformationError(msg)
+
+    @dilution.setter
+    def dilution(self, value):
+        self._sample_metadata.loc[self._sample_mask, _sample_dilution] = value
 
     def is_valid_class_name(self, test_class: Union[str, List[str]]) -> bool:
         """
@@ -316,6 +286,7 @@ class DataContainer(object):
         Returns
         -------
         is_valid : bool
+
         """
         valid_classes = self.classes.unique()
         if isinstance(test_class, str):
@@ -335,6 +306,7 @@ class DataContainer(object):
         remove : Iterable[str]
             List of sample/feature names to remove.
         axis : {"features", "samples"}
+
         """
         
         if not self._is_valid(remove, axis):
@@ -355,6 +327,7 @@ class DataContainer(object):
         index: list[str]
             List of feature/sample names to check.
         axis: {"samples", "features"}
+
         """
         ind = pd.Index(index)
         if axis == "features":
@@ -378,6 +351,7 @@ class DataContainer(object):
             is True if there are NANs in the data matrix;  "order" is True
             if there is run order information for the samples; "batch" is True
             if there is batch number information associated to the samples.
+
         """
         
         diagnostic = dict()
@@ -386,6 +360,7 @@ class DataContainer(object):
         diagnostic[_qc_sample_type] = bool(self.mapping[_qc_sample_type])
         diagnostic[_blank_sample_type] = bool(self.mapping[_blank_sample_type])
         diagnostic[_study_sample_type] = bool(self.mapping[_study_sample_type])
+        diagnostic[_dilution_qc_type] = bool(self.mapping[_dilution_qc_type])
         try:
             diagnostic[_sample_order] = self.order.any()
         except RunOrderError:
@@ -414,21 +389,6 @@ class DataContainer(object):
         if reset_mapping:
             self.mapping = None
 
-    # TODO: test and remove
-    # def get_mapped_classes(self) -> List[str]:
-    #     """
-    #     return all classes assigned in the sample type mapping.
-    #
-    #     Returns
-    #     -------
-    #     process_classes: list of mapped classes.
-    #     """
-    #     process_classes = list()
-    #     for classes in self.mapping.values():
-    #         if classes is not None:
-    #             process_classes += classes
-    #     return process_classes
-
     def select_features(self, mzq: float, rtq: float, mz_tol: float = 0.01,
                         rt_tol: float = 5) -> pd.Index:
         """
@@ -605,80 +565,68 @@ class MetricMethods:
 
     Methods
     -------
-    cv(intraclass=True, robust=False): Computes the coefficient of variation for
-    each feature.
-    dratio(robust=False, sample_classes=None, qc_classes=None): Computes the
-    D-Ratio of features, a metric used to compare technical to biological
-    variation.
-    detection_rate(intraclass=True, threshold=0): Computes the ratio of samples
-    where a features was detected.
-    pca(n_components=2, normalization=None, scaling=None): Computes the PCA
-    scores, loadings and  PC variance.
+    cv: Computes the coefficient of variation for each feature.
+    dratio: Computes the D-Ratio of features, a metric used to compare technical
+    to biological variation.
+    detection_rate: Computes the ratio of samples where a features was detected.
+    pca: Computes the PCA scores, loadings and  PC variance.
 
     """
     
     def __init__(self, data: DataContainer):
         self.__data = data
     
-    def cv(self, intraclass: bool = True, robust: bool = False,
-           fill_value: float = np.inf):
+    def cv(self, groupby: Union[str, List[str], None] = "class",
+           robust: bool = False, fill_value: float = np.inf):
         """
-        Computes the Coefficient of variation defined as the ratio between the
-        standard deviation and the mean of each feature.
+        Computes the Coefficient of variation for each feature.
+
+        The coefficient of variation is the quotient between the standard
+        deviation and the mean of a feature.
         
         Parameters
         ----------
-        intraclass: True
-            if True computes the coefficient of variation for each
-            class. Else computes the mean coefficient of variation
-            for all the classes mapped as `sample` type.
+        groupby: str, List[str] or None
+            If groupby isa column or a list of columns of sample metadata, the
+            values of CV are computed on a per group basis. If None, the CV
+            is computed for all samples in the data.
         robust: bool
             If True, computes the relative MAD. Else, computes the Coefficient
             of variation.
         fill_value: float
-            Value used to replace NaN. np.inf is used to facilitate features
-            with NaN values.
+            Value used to replace NaN. By default, NaNs are replaced by np.inf.
 
         Returns
         -------
-        result: pd.Series or pd.DataFrame
+        pd.Series or pd.DataFrame
+
         """
         if robust:
             cv_func = utils.robust_cv
         else:
             cv_func = utils.cv
 
-        # fill value is np.inf to facilitate filtering NaN features
-        if intraclass:
-            result = (self.__data.data_matrix
-                      .groupby(self.__data.classes)
+        if groupby is not None:
+            if isinstance(groupby, str):
+                by = self.__data.sample_metadata[groupby]
+            else:
+                by = [self.__data.sample_metadata[x] for x in groupby]
+            result = (self.__data.data_matrix.groupby(by)
                       .apply(cv_func, fill_value=fill_value))
         else:
-            if self.__data.mapping[_study_sample_type] is None:
-                result = cv_func(self.__data.data_matrix, fill_value=fill_value)
-            else:
-                sample_class = self.__data.mapping[_study_sample_type]
-                is_sample_class = self.__data.classes.isin(sample_class)
-                result = cv_func(self.__data.data_matrix[is_sample_class],
-                                 fill_value=fill_value)
+            result = cv_func(self.__data.data_matrix, fill_value=fill_value)
         return result
-    
-    def dratio(self, robust=False,
-               sample_classes: Optional[List[str]] = None,
-               qc_classes: Optional[List[str]] = None) -> pd.Series:
+
+    def dratio(self, robust=False) -> pd.Series:
         """
-        Computes the D-Ratio using sample variation and quality control
-        variation [BD18]. This metric is useful to compare technical to biological
+        Computes the ratio between the sample variation and the quality control
         variation.
+
+        The D-Ratio is useful to compare technical to biological variation and
+        non informative features.
         
         Parameters
         ----------
-        sample_classes: list[str], optional
-            classes used to estimate biological variation. If None, uses
-            values from sample_type in sample mapping
-        qc_classes: list[str], optional
-            classes used to estimate technical variation. If None, uses
-            values from qc_type in sample mapping
         robust: bool
             If True, uses MAD to compute the D-ratio. Else, uses standard
             deviation.
@@ -688,19 +636,16 @@ class MetricMethods:
         dr : pd.Series:
             D-Ratio for each feature
 
-        References
-        ----------
-        .. [BD18] D.Broadhurst *et al*, "Guidelines and considerations for the use
-            of system suitability and quality control samples in mass spectrometry
-            assays applied in untargeted clinical metabolomic studies",
-            Metabolomics (2018) 14:72.
-
         """
-        if sample_classes is None:
-            sample_classes = self.__data.mapping[_study_sample_type]
+        if self.__data.mapping[_study_sample_type] is None:
+            msg = "Study sample classes not specified in the sample mapping"
+            raise ValueError(msg)
+        sample_classes = self.__data.mapping[_study_sample_type]
 
-        if qc_classes is None:
-            qc_classes = self.__data.mapping[_qc_sample_type]
+        if self.__data.mapping[_qc_sample_type] is None:
+            msg = "QC class not specified in the sample mapping"
+            raise ValueError(msg)
+        qc_classes = self.__data.mapping[_qc_sample_type]
 
         is_sample_class = self.__data.classes.isin(sample_classes)
         is_qc_class = self.__data.classes.isin(qc_classes)
@@ -711,30 +656,33 @@ class MetricMethods:
                                 fill_value=np.inf)
         return dratio
     
-    def detection_rate(self, intraclass=True, threshold=0):
+    def detection_rate(self, groupby: Union[str, List[str], None] = "class",
+                       threshold=0):
         """
         Computes the fraction of samples where a feature was detected.
         
         Parameters
         ----------
-        intraclass: bool
-            if True, computes the detection rate for each class, else
-            computes the mean detection rate
-        threshold: float
+        groupby : str, List[str] or None
+            If groupby isa column or a list of columns of sample metadata, the
+            values of CV are computed on a per group basis. If None, the CV
+            is computed for all samples in the data.
+        threshold : float
             Minimum value to consider a feature detected
-        """
 
-        if intraclass:
-            results = (self.__data.data_matrix
-                       .groupby(self.__data.classes)
-                       .apply(utils.detection_rate, threshold=threshold))
+        """
+        if groupby is not None:
+            if isinstance(groupby, str):
+                by = self.__data.sample_metadata[groupby]
+            else:
+                by = [self.__data.sample_metadata[x] for x in groupby]
+            result = (self.__data.data_matrix.groupby(by)
+                      .apply(utils.detection_rate, threshold=threshold))
         else:
-            # TODO: same as CV. check if its better to return a global DR.
-            sample_class = self.__data.mapping[_study_sample_type]
-            is_sample_class = self.__data.classes.isin(sample_class)
-            results = self.__data.data_matrix[is_sample_class].apply()
-        return results
-    
+            result = (self.__data.data_matrix
+                      .apply(utils.detection_rate, threshold=threshold))
+        return result
+
     def pca(self, n_components: Optional[int] = 2,
             normalization: Optional[str] = None,
             scaling: Optional[str] = None,
@@ -770,20 +718,19 @@ class MetricMethods:
         else:
             ind = data.index
 
-        if normalization:
+        if normalization is not None:
             data = utils.normalize(data, normalization)
 
-        if scaling:
+        if scaling is not None:
             data = utils.scale(data, scaling)
 
         pca = PCA(n_components=n_components)
         scores = pca.fit_transform(data)
+        n_components = pca.n_components_    # effective number of pc
         loadings = pca.components_.T * np.sqrt(pca.explained_variance_)
         variance = pca.explained_variance_
         pc_str = ["PC" + str(x) for x in range(1, n_components + 1)]
-        scores = pd.DataFrame(data=scores,
-                              index=ind,
-                              columns=pc_str)
+        scores = pd.DataFrame(data=scores, index=ind, columns=pc_str)
         loadings = pd.DataFrame(data=loadings,
                                 index=self.__data.data_matrix.columns,
                                 columns=pc_str)
@@ -791,8 +738,46 @@ class MetricMethods:
         total_variance = data.var().sum()
         return scores, loadings, variance, total_variance
 
+    def correlation(self, field: str, mode: str = "ols",
+                    classes: Optional[List[str]] = None):
+        """
+        Correlates features with sample metadata properties.
 
-class BokehPlotMethods:
+        Parameters
+        ----------
+        field : str
+            A column of `sample_metadata`. Must have a numeric dtype.
+        mode: {"ols", "spearman"}
+            `ols` computes the ordinary least squares linear regression.
+            Computes the Pearson r squared, p-value for the Jarque-Bera test and
+            the Durwin-Watson statistic for each feature. `spearman` computes
+            the spearman rank correlation coefficient for each feature
+        classes: List[str], optional
+            Compute the correlation on the selected classes only. If None,
+            computes the correlation on all samples.
+
+        Returns
+        -------
+        pandas.Series or pandas.DataFrame
+
+        """
+        if mode not in ["ols", "spearman"]:
+            msg = "Valid modes are `ols` or `spearman`"
+            raise ValueError(msg)
+
+        if classes is None:
+            x = self.__data.sample_metadata[field]
+            df = self.__data.data_matrix
+        else:
+            mask = self.__data.classes.isin(classes)
+            x = self.__data.sample_metadata.loc[mask, field]
+            df = self.__data.data_matrix[mask]
+        correlation_aux = lambda y: utils.metadata_correlation(y, x, mode)
+        res = df.apply(correlation_aux, result_type="expand")
+        return res
+
+
+class BokehPlotMethods:  # pragma: no cover
     """
     Methods to plot data from a DataContainer. Generates Bokeh Figures.
 
@@ -877,7 +862,8 @@ class BokehPlotMethods:
 
         if hue == _sample_type:
             rev_map = _reverse_mapping(self._data_container.mapping)
-            score[_sample_type] = score[_sample_class].apply(lambda x: rev_map.get(x))
+            score[_sample_type] = (score[_sample_class]
+                                   .apply(lambda x: rev_map.get(x)))
             score = score[~pd.isna(score[_sample_type])]
         elif hue == _sample_batch:
             score[_sample_batch] = score[_sample_batch].astype(str)
@@ -1034,7 +1020,7 @@ class BokehPlotMethods:
         if hue == _sample_type:
             rev_map = _reverse_mapping(self._data_container.mapping)
             source[_sample_type] = (source[_sample_class]
-                              .apply(lambda x: rev_map.get(x)))
+                                    .apply(lambda x: rev_map.get(x)))
             source = source[~source[_sample_type].isna()]
         elif hue == _sample_batch:
             source[_sample_batch] = source[_sample_batch].astype(str)
@@ -1068,7 +1054,7 @@ class BokehPlotMethods:
         return fig
 
 
-class SeabornPlotMethods(object):
+class SeabornPlotMethods(object):   # pragma: no cover
     """
     Methods to plot feature data from a DataContainer using Matplotlib/Seaborn.
     """
@@ -1130,7 +1116,8 @@ class SeabornPlotMethods(object):
 
         if hue == _sample_type:
             rev_map = _reverse_mapping(self._data.mapping)
-            score[_sample_type] = score[_sample_class].apply(lambda s: rev_map.get(s))
+            score[_sample_type] = (score[_sample_class]
+                                   .apply(lambda s: rev_map.get(s)))
             score = score[~pd.isna(score[_sample_type])]
         elif hue == _sample_batch:
             score[_sample_batch] = score[_sample_batch].astype(str)
@@ -1143,7 +1130,7 @@ class SeabornPlotMethods(object):
             for ind in score.index:
                 x = score.loc[ind, x_name] * 1.01
                 y = score.loc[ind, y_name] * 1.01
-                t = str(self._data.order[ind])
+                t = str(self._data.order[ind].astype(int))
                 g.ax.annotate(t, (x, y))
 
         # set x and y label
@@ -1238,6 +1225,7 @@ class SeabornPlotMethods(object):
             df[_sample_class] = df[_sample_class].astype(str)
 
         g = sns.relplot(data=df, **relplot_params)
+        return g
 
 
 class PreprocessMethods:
@@ -1361,6 +1349,13 @@ class EmptyDataContainerError(Exception):
     pass
 
 
+class DilutionInformationError(Exception):
+    """
+    Error class raised when no dilution factor information has been provided.
+    """
+    pass
+
+
 def _validate_mapping(mapping, valid_samples):
     for sample_type, classes in mapping.items():
         if sample_type not in SAMPLE_TYPES:
@@ -1419,21 +1414,6 @@ def _make_empty_mapping():
     return empty_mapping
 
 
-# def _validate_batch_order(batch: pd.Series, order: pd.Series):
-#     if batch.dtype != int:
-#         msg = "batch must be of integer dtype"
-#         raise BatchInformationError(msg)
-#     if order.dtype != int:
-#         msg = "order must be of integer dtype"
-#         raise RunOrderError(msg)
-#
-#     grouped = order.groupby(batch)
-#     for _, batch_order in grouped:
-#         if not np.array_equal(batch_order, batch_order.unique()):
-#             msg = "order value must be unique for each batch"
-#             raise RunOrderError(msg)
-
-
 def _reverse_mapping(mapping):
     rev_map = dict()
     for k, v in mapping.items():
@@ -1441,7 +1421,3 @@ def _reverse_mapping(mapping):
             for class_value in v:
                 rev_map[class_value] = k
     return rev_map
-
-
-# TODO: subclass DataContainer into  NMRDataContainer and MSDataContainer
-#  (add DI, LC)
