@@ -78,7 +78,9 @@ class Reporter(object):
         status: {"before", "after"}
         """
         metrics = dict()
-        metrics["cv"] = dc.metrics.cv(intraclass=False, fill_value=0).mean()
+        metrics["cv"] = (dc.metrics.cv(groupby="class", fill_value=0)
+                         .median()
+                         .median())
         n_samples, n_features = dc.data_matrix.shape
         metrics["features"] = n_features
         metrics["samples"] = n_samples
@@ -303,9 +305,17 @@ class DuplicateMerger(Processor):
 class ClassRemover(Processor):
     """
     Remove samples from the specified classes.
+
+    Parameters
+    ----------
+    classes: list
+        List of classes to remove.
+
     """
     def __init__(self, classes: List[str]):
-        super(ClassRemover, self).__init__(axis="samples", mode="filter")
+        requirements = {"empty": False, "missing": False}
+        super(ClassRemover, self).__init__(axis="samples", mode="filter",
+                                           requirements=requirements)
         self.name = "Class Remover"
         if classes is None:
             self.params["classes"] = list()
@@ -372,12 +382,14 @@ class BlankCorrector(Processor):
     def __init__(self, corrector_classes: Optional[List[str]] = None,
                  process_classes: Optional[List[str]] = None,
                  mode: Union[str, Callable] = "lod", factor: float = 1,
+                 robust: bool = True,
                  process_blanks: bool = True, verbose=False):
         """
         Constructor for the BlankCorrector.
 
         """
-        requirements = {"empty": False, "missing": False}
+        requirements = {"empty": False, "missing": False,
+                        _blank_sample_type: True}
         super(BlankCorrector, self).__init__(axis=None, mode="transform",
                                              verbose=verbose,
                                              requirements=requirements)
@@ -387,6 +399,7 @@ class BlankCorrector(Processor):
         self.params["mode"] = mode
         self.params["factor"] = factor
         self.params["process_blanks"] = process_blanks
+        self.params["robust"] = robust
         self._default_process = _study_sample_type
         self._default_correct = _blank_sample_type
         validation.validate_blank_corrector_params(self.params)
@@ -452,7 +465,11 @@ class PrevalenceFilter(Processor):
         validation.validate_prevalence_filter_params(self.params)
 
     def func(self, dc):
-        dr = dc.metrics.detection_rate(intraclass=self.params["intraclass"],
+        if self.params["intraclass"]:
+            groupby = "class"
+        else:
+            groupby = None
+        dr = dc.metrics.detection_rate(groupby=groupby,
                                        threshold=self.params["threshold"])
         dr = dr.loc[self.params["process_classes"], :]
         lb = self.params["lb"]
@@ -580,14 +597,87 @@ class VariationFilter(Processor):
     def func(self, dc: DataContainer):
         lb = self.params["lb"]
         ub = self.params["ub"]
-        variation = dc.metrics.cv(intraclass=self.params["intraclass"],
-                                  robust=self.params["robust"])
+        if self.params["intraclass"]:
+            groupby = "class"
+        else:
+            groupby = None
+        variation = dc.metrics.cv(groupby=groupby, robust=self.params["robust"])
         if self.params["intraclass"]:
             variation = variation.loc[self.params["process_classes"], :]
 
         return get_outside_bounds_index(variation, lb, ub)
 
 
+@register
+class DilutionFilter(Processor):
+    """
+    Filter features based on the correlation with a dilution factor.
+
+    In order to use this filter, the dilution column must be specified in the
+    sample_metadata of the DataContainer. Also, the QCs used for the
+    analysis must be specified under the `dqc` key in the DataContainer mapping.
+
+    Parameters
+    ----------
+    min_corr : number between 0 and 1
+        Lower bound for the correlation coefficient.
+    plim : number between 0 and 1
+        p-value limit for the Jarque-Bera test. Used only when mode is `ols`.
+    mode : {"ols", "spearman"}
+        `ols` computes the ordinary least squares linear regression.The r
+        squared from the fit and the p-value from the Jarque-Bera test are used
+        to evaluate the linearity of the signal with the dilution. Features with
+        correlation values lower than `min_corr` or p-values lower than `plim`
+        are removed. `spearman` compares the correlation threshold with the
+        spearman rank correlation coefficient.
+    verbose: bool
+        If True, prints a message
+
+    Notes
+    -----
+    Correlation with the dilution is a measure of the linearity of the response
+    of the feature in the experimental conditions [2].
+
+    References
+    ----------
+    ..  [2] Lewis MR, *et al*, Development and Application of Ultra-Performance
+        Liquid Chromatography-TOF MS for Precision Large Scale Urinary Metabolic
+        Phenotyping. Anal Chem. (2016), 88(18):9004-13.
+        doi: 10.1021/acs.analchem.6b01481
+
+    """
+    def __init__(self, min_corr: float = 0.8, plim: float = 0.1,
+                 mode: str = "ols", verbose: bool = False):
+        """
+        Constructor of the DilutionFilter.
+        """
+        requirements = {"empty": False, "missing": False,
+                        _dilution_qc_type: True}
+        super(DilutionFilter, self).__init__(axis="features", mode="filter",
+                                             verbose=verbose,
+                                             requirements=requirements)
+        self.name = "Dilution Filter"
+        self.params["min_corr"] = min_corr
+        self.params["plim"] = plim
+        self.params["mode"] = mode
+        self._default_process = _qc_sample_type
+        validation.validate_dilution_filter_params(self.params)
+
+    def func(self, dc: DataContainer):
+        mode = self.params["mode"]
+        min_corr = self.params["min_corr"]
+        plim = self.params["plim"]
+        corr = dc.metrics.correlation("dilution", mode=mode,
+                                      classes=dc.mapping[_dilution_qc_type])
+        if mode == "ols":
+            r2_ind = get_outside_bounds_index(corr.loc["r2", :], min_corr, 1)
+            jb_ind = get_outside_bounds_index(corr.loc["JB", :], plim, 1)
+            rm_ind = r2_ind.union(jb_ind)
+        else:
+            rm_ind = get_outside_bounds_index(corr, min_corr, 1)
+        return rm_ind
+
+
 class _TemplateValidationFilter(Processor):
     """
     Checks that process samples in each batch are surrounded by corrector
@@ -734,6 +824,7 @@ class _BatchCorrectorProcessor(Processor):
                  frac: Optional[float] = None,
                  interpolator: str = "splines",
                  n_qc: Optional[int] = None,
+                 method: str = "multiplicative",
                  process_qc: bool = True,
                  verbose: bool = False):
         requirements = {"empty": False, "missing": False, _sample_order: True,
@@ -749,6 +840,7 @@ class _BatchCorrectorProcessor(Processor):
         self.params["interpolator"] = interpolator
         self.params["n_qc"] = n_qc
         self.params["frac"] = frac
+        self.params["method"] = method
         self.params["process_qc"] = process_qc
         self._default_process = _study_sample_type
         self._default_correct = _qc_sample_type
@@ -783,12 +875,16 @@ class BatchCorrector(Pipeline):
         feature is estimated using LOOCV.
     interpolator : {"splines", "linear"}
         Interpolator used to estimate the correction for each sample.
+    method : {"additive", "multiplicative"}
+        Method used to model the variation in samples.
     corrector_classes : list[str], optional
         list of classes used to generate the correction. If None uses
         QC sample types from the mapping.
     process_classes : list[str], optional
         list of classes used to correct. If None uses
         sample sample types from the mapping.
+    scale_blanks : bool
+        If True, scale blanks across batches.
     verbose : bool
         If True a message is shown after processing the data matrix.
 
@@ -813,7 +909,7 @@ class BatchCorrector(Pipeline):
 
     In order to apply this correction, several checks needs to be made. First,
     the QC template is checked and samples that cannot be corrected are removed.
-    A study sample is valid if it is surrounded by a quality control sample.
+    A study sample is valid if it is surrounded by QC samples.
     This is a necessary step because the correction for the study samples is
     built using interpolation. It's recommended to have three QC samples at the
     beginning and at the end of each batch. See [1] for recommendations on
@@ -860,7 +956,7 @@ class BatchCorrector(Pipeline):
     After these two checks, the remaining samples and features are suitable
     for LOESS batch correction. A final consideration is how to estimate the
     :math:`\bar{m_{k}}` for each feature. This value is usually computed as the
-    mean or median of the QC values in a batch, but ff the temporal bias becomes
+    mean or median of the QC values in a batch, but if the temporal bias becomes
     stronger as more samples are analyzed, a better estimation of
     :math:`\bar{m_{k}}` can be obtained using the average of the first samples
     analyzed in a batch. To this end, the `n_qc` parameter controls how many
@@ -879,8 +975,10 @@ class BatchCorrector(Pipeline):
                  first_n_qc: Optional[int] = None,
                  threshold: float = 0, frac: Optional[float] = None,
                  interpolator: str = "splines",
+                 method: str = "multiplicative",
                  corrector_classes: Optional[List[str]] = None,
                  process_classes: Optional[List[str]] = None,
+                 scale_blanks: bool = True,
                  verbose: bool = False):
 
         checker = \
@@ -897,7 +995,7 @@ class BatchCorrector(Pipeline):
             _BatchCorrectorProcessor(corrector_classes=corrector_classes,
                                      process_classes=process_classes, frac=frac,
                                      interpolator=interpolator, verbose=verbose,
-                                     n_qc=first_n_qc)
+                                     n_qc=first_n_qc, method=method)
         pipeline = [checker, prevalence, corrector]
         super(BatchCorrector, self).__init__(pipeline, verbose=verbose)
         self.name = "Batch Corrector"
