@@ -24,9 +24,15 @@ find_closest(x, xq) : Finds the elements in xq closest to x.
 
 import numpy as np
 import pandas as pd
+from statsmodels.api import OLS, add_constant
+from statsmodels.stats.stattools import jarque_bera, durbin_watson
+from scipy.stats import spearmanr, median_abs_deviation
 import os.path
 from typing import Optional, Union
 
+data_type = Union[pd.DataFrame, pd.Series]
+reduced_type = Union[pd.Series, float]
+
 
 def gauss(x: np.ndarray, mu: float, sigma: float,
           amp: float):  # pragma: no cover
@@ -181,7 +187,7 @@ def sample_to_path(samples, path):
     d : dict
 
     """
-    # TODO: this function should accept and extension parameter to prevent
+    # TODO: this function should accept an extension parameter to prevent
     #   files with the same name but invalid extensions from being used
     available_files = os.listdir(path)
     filenames = [os.path.splitext(x)[0] for x in available_files]
@@ -193,7 +199,7 @@ def sample_to_path(samples, path):
     return d
 
     
-def cv(df: pd.DataFrame, fill_value: Optional[float] = None) -> pd.Series:
+def cv(df: data_type, fill_value: Optional[float] = None) -> reduced_type:
     """
     Computes the Coefficient of variation for each column.
 
@@ -201,12 +207,12 @@ def cv(df: pd.DataFrame, fill_value: Optional[float] = None) -> pd.Series:
 
     """
     res = df.std() / df.mean()
-    if fill_value is not None:
-        res = res.fillna(fill_value)
+    res = _fill_na(res, fill_value)
     return res
 
 
-def robust_cv(df, fill_value: Optional[float] = None):
+def robust_cv(df: data_type, fill_value: Optional[float] = None
+              ) -> reduced_type:
     """
     Estimation of the coefficient of variation using the MAD and median.
     Assumes a normal distribution.
@@ -214,19 +220,25 @@ def robust_cv(df, fill_value: Optional[float] = None):
 
     # 1.4826 is used to estimate sigma in an unbiased way assuming a normal
     # distribution for each feature.
-    res = 1.4826 * df.mad() / df.median()
-    if fill_value is not None:
-        res = res.fillna(fill_value)
+    res = mad(df) / df.median()
+    res = _fill_na(res, fill_value)
     return res
 
 
-def mad(df):
+def mad(df: data_type) -> reduced_type:
     """
     Computes the median absolute deviation for each column. Fill missing
     values with zero.
     """
-    res = df.mad()
-    res = res.fillna(0)
+    # for dataframes with only one row a series of nan is returned. This is
+    # to return the same value as std.
+    if isinstance(df, pd.DataFrame):
+        if df.shape[0] == 1:
+            res = pd.Series(data=np.nan, index=df.columns)
+        else:
+            res = df.apply(median_abs_deviation, scale="normal")
+    else:
+        res = median_abs_deviation(df, scale="normal")
     return res
 
 
@@ -236,7 +248,7 @@ def sd_ratio(df1: pd.DataFrame, df2: pd.DataFrame, robust: bool = False,
     Computes the ratio between the standard deviation of the columns of
     DataFrame1 and DataFrame2.
 
-    Used to compute the D-Ratio metric. NaN values are filled to np.inf.
+    Used to compute the D-Ratio metric.
 
     Parameters
     ----------
@@ -256,13 +268,11 @@ def sd_ratio(df1: pd.DataFrame, df2: pd.DataFrame, robust: bool = False,
         ratio = mad(df1) / mad(df2)
     else:
         ratio = df1.std() / df2.std()
-
-    if fill_value is not None:
-        ratio = ratio.fillna(fill_value)
+    ratio = _fill_na(ratio, fill_value)
     return ratio
 
 
-def detection_rate(df: pd.DataFrame, threshold: float = 0.0) -> pd.Series:
+def detection_rate(df: data_type, threshold: float = 0.0) -> reduced_type:
     """
     Computes the fraction of values in a column above the `threshold`.
 
@@ -276,11 +286,56 @@ def detection_rate(df: pd.DataFrame, threshold: float = 0.0) -> pd.Series:
     dr : pd.Series
 
     """
-    # dr = df[df > threshold].count() / df.count()
-    dr = (df > threshold).sum().astype(int) / df.shape[0]
+    if isinstance(df, pd.DataFrame):
+        n = df.shape[0]
+    else:
+        n = df.size
+
+    dr = (df > threshold).sum().astype(int) / n
     return dr
 
 
+def metadata_correlation(y, x, mode: str = "ols"):
+    """
+    Computes correlation metrics between two variables.
+
+    Parameters
+    ----------
+    y : array
+    x : array
+    mode: {"ols", "spearman"}
+        `ols` computes r squared, Jarque-Bera test p-value and Durwin-Watson
+        statistic from the ordinary least squares linear regression. `spearman`
+        computes the spearman rank correlation coefficient.
+
+    Returns
+    -------
+    dict
+
+    """
+    if mode == "ols":
+        ols = OLS(y, add_constant(x)).fit()
+        r2 = ols.rsquared
+        jb = jarque_bera(ols.resid)[1]  # Jarque Bera test p-value
+        dw = durbin_watson(ols.resid)   # Durwin Watson statistic
+        res = {"r2": r2, "DW": dw, "JB": jb}
+    else:
+        res = spearmanr(y, x)[0]
+    return res
+
+
+def _fill_na(s: reduced_type, fill_value: Optional[float]):
+    if fill_value is None:
+        res = s
+    elif isinstance(s, pd.Series):
+        res = s.fillna(fill_value)
+    elif pd.isna(s):
+        res = fill_value
+    else:
+        res = s
+    return res
+
+
 def _find_closest_sorted(x: np.ndarray,
                          xq: Union[np.ndarray, float, int]) -> np.ndarray:
     """
@@ -362,23 +417,30 @@ def is_unique(s: pd.Series):
 class SimulatedExperiment:  # pragma: no cover
     """
     Emulates a pyopenms MSExperiment. Used for tests.
+
     """
     def __init__(self, mz_values: np.ndarray, rt_values: np.ndarray,
                  mz_params: np.ndarray, rt_params: np.ndarray,
-                 noise: Optional[float] = None):
+                 ft_noise: Optional[np.ndarray] = None,
+                 noise: Optional[float] = None, mode: str = "centroid"):
         """
         Constructor function
 
         Parameters
         ----------
-        mz_values : sorted mz values for each mz scan
-        rt_values : sorted rt values
+        mz_values : array
+            sorted mz values for each mz scan, used to build spectra in profile
+            mode.
+        rt_values : array
+            sorted rt values, used to set scan time.
         mz_params : array with shape (n, 3)
              Used to build m/z peaks. Each row is the mean, standard deviation
-             and amplitude in the m/z dimension
+             and amplitude in the m/z dimension.
         rt_params : array with shape (n, 3)
              Used to build rt peaks. Each row is the mean, standard deviation
              and amplitude in the rt dimension
+        ft_noise : array_with shape (n, 2), optional
+            adds noise to mz and rt values
         noise : positive number, optional
             noise level to add to each scan. the noise is modeled as gaussian
             iid noise in each scan with standard deviation equal to the value
@@ -387,18 +449,26 @@ class SimulatedExperiment:  # pragma: no cover
             a seed value associated to generate always the same noise value in
             a given scan. seed values are generated randomly each time a new
             object is instantiated.
+        mode : {"centroid", "profile"}
 
         """
         self.mz = mz_values
-        self.mz_array = gaussian_mixture(mz_values, mz_params)
+        self.mz_params = mz_params
         self.rt = rt_values
-        self.rt_array = gaussian_mixture(rt_values, rt_params)
         self.n_scans = rt_values.size
         self._seeds = None
         self._noise_level = None
-        if noise is not None:
-            self._seeds = np.random.choice(100000, self.n_scans)
-            self._noise_level = noise
+        self.ft_noise = ft_noise
+        self.mode = mode
+        # seeds are used to ensure that each time that a spectra is generated
+        # with get_spectra its values are the same
+        self._seeds = np.random.choice(self.n_scans * 10, self.n_scans)
+        self._noise_level = noise
+
+        if ft_noise is not None:
+            np.random.seed(self._seeds[0])
+            rt_params[0] += np.random.normal(scale=ft_noise[1])
+        self.rt_array = gaussian_mixture(rt_values, rt_params)
 
     def getNrSpectra(self):
         return self.n_scans
@@ -409,14 +479,31 @@ class SimulatedExperiment:  # pragma: no cover
             msg = "Invalid scan number."
             raise ValueError(msg)
         rt = self.rt[scan_number]
-        spint = self.rt_array[:, scan_number][:, np.newaxis] * self.mz_array
-        spint = spint.sum(axis=0)
+
+        # create a mz array for the current scan
+        if self.ft_noise is not None:
+            np.random.seed(self._seeds[scan_number])
+            mz_noise = np.random.normal(scale=self.ft_noise[0])
+            mz_params = self.mz_params.copy()
+            mz_params[:, 0] += mz_noise
+        else:
+            mz_params = self.mz_params
+
+        if self.mode == "centroid":
+            mz = mz_params[:, 0]
+            spint = self.rt_array[:, scan_number] * mz_params[:, 1]
+        else:
+            mz = self.mz
+            mz_array = gaussian_mixture(self.mz, mz_params)
+            spint = self.rt_array[:, scan_number][:, np.newaxis] * mz_array
+            spint = spint.sum(axis=0)
+
         if self._noise_level is not None:
             np.random.seed(self._seeds[scan_number])
-            noise = np.random.normal(size=self.mz.size, scale=self._noise_level)
+            noise = np.random.normal(size=mz.size, scale=self._noise_level)
             noise -= noise.min()
             spint += noise
-        sp = SimulatedSpectrum(self.mz, spint, rt)
+        sp = SimulatedSpectrum(mz, spint, rt)
         return sp
 
 
@@ -434,3 +521,7 @@ class SimulatedSpectrum:    # pragma: no cover
 
     def get_peaks(self):
         return self.mz, self.spint
+
+    def getMSLevel(self):
+        return 1
+
